﻿using System;
using System.IO;
using System.Diagnostics;
using System.Drawing;

namespace s4_visualizer
{
    class Executor
    {
        public Executor(Log _log)
        {
            workDir = "wkdir";
            exeDest = workDir + "\\_sol.exec";
            inputDest = workDir + "\\flow.in";
            outputDest = workDir + "\\flow.out";

            generatorExe = "generator.exe";
            inCheckerExe = "input-checker.exe";
            outCheckerExe = "output-checker.exe";

            log = _log;
        }

        public bool runSolution (string exeFile, string command)
        {
            // delete the output file left from a previous execution
            if (File.Exists(outputDest))
            {
                File.Delete(outputDest);
            }

            if (!File.Exists(exeFile))
            {
                log.print(string.Format("Error: executable file \"{0}\" not found.\n", exeFile), Color.Red);
                return false;
            }
            else
            {
                if (File.Exists(exeDest))
                {
                    File.Delete(exeDest);
                }
                File.Copy(exeFile, exeDest);
                log.print(string.Format("Copied \"{0}\" to \".\\{1}\"\n", exeFile, exeDest));
            }

            Process proc = new Process();
            proc.StartInfo.FileName = "cmd.exe";
            proc.StartInfo.WorkingDirectory = workDir;
            proc.StartInfo.Arguments = string.Format(" /C {0}", command);
            proc.StartInfo.UseShellExecute = false;

            log.print("Running solution...   ");
            proc.Start();

            proc.WaitForExit();

            if (proc.ExitCode == 0)
            {
                log.print("OK\n");
                return true;
            }
            else
            {
                log.print(string.Format("\nError: Process exited with exit code 0x{0:X}\n", proc.ExitCode), Color.Red);
                return false;
            }
        }

        public bool generate (ref int seed, int gridSz, int nFlows, int nBlocked)
        {
            if (seed == -1)
            {
                Random rnd = new Random((int)DateTime.Now.Ticks);
                seed = rnd.Next();
            }

            Process proc = new Process();
            proc.StartInfo.FileName = "cmd.exe";
            proc.StartInfo.WorkingDirectory = workDir;
            proc.StartInfo.Arguments = string.Format("/C {0} {1} {2} {3} {4}", 
                new object[]{generatorExe, seed, gridSz, nFlows, nBlocked});
            proc.StartInfo.UseShellExecute = false;

            log.print(string.Format("Generating input file with seed {0}...   ", seed));
            proc.Start();

            proc.WaitForExit();

            if (proc.ExitCode == 0)
            {
                log.print("Done\n");
                return true;
            }
            else
            {
                log.print("\nError: Could not generate input file (Maybe you used incompatible arguments)\n", Color.Red);
                return false;
            }
        }

        public bool copyAndCheckInput (string inputFile)
        {
            if (!File.Exists(inputFile))
            {
                log.print(string.Format("Error: Input file \"{0}\" not found.\n", inputFile), Color.Red);
                return false;
            }
            else
            {
                if (File.Exists(inputDest))
                {
                    File.Delete(inputDest);
                }
                File.Copy(inputFile, inputDest);
                log.print(string.Format("Copied \"{0}\" to \".\\{1}\"\n", inputFile, inputDest));
            }

            Process proc = new Process();
            proc.StartInfo.FileName = "cmd.exe";
            proc.StartInfo.WorkingDirectory = workDir;
            proc.StartInfo.UseShellExecute = false;
            proc.StartInfo.Arguments = string.Format("/C {0}", inCheckerExe);
            proc.StartInfo.RedirectStandardOutput = true;

            log.print("Checking input file...   ");
            proc.Start();

            proc.WaitForExit();

            if (proc.ExitCode == 0)
            {
                log.print("OK\n");
                return true;
            }
            else
            {
                string errMsg = proc.StandardOutput.ReadToEnd();
                log.print("\n" + errMsg, Color.Red);
                return false;
            }
        }

        public bool checkOutput ()
        {
            Process proc = new Process();
            proc.StartInfo.FileName = "cmd.exe";
            proc.StartInfo.WorkingDirectory = workDir;
            proc.StartInfo.UseShellExecute = false;
            proc.StartInfo.Arguments = string.Format("/C {0}", outCheckerExe);
            proc.StartInfo.RedirectStandardOutput = true;

            log.print("Checking output file...   ");
            proc.Start();

            proc.WaitForExit();

            if (proc.ExitCode == 0)
            {
                log.print("OK\n");
                return true;
            }
            else
            {
                string errMsg = proc.StandardOutput.ReadToEnd();
                log.print("\n" + errMsg, Color.Red);
                return false;
            }
        }

        public string workDir;
        public string exeDest;
        public string inputDest;
        public string outputDest;

        private string generatorExe;
        private string inCheckerExe;
        private string outCheckerExe;

        Log log;
    }
}