﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace s4_visualizer
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();

            log = new Log(logBox);
            executor = new Executor(log);
            renderer = new Renderer(renderSurface);
            simulator = new Simulator(renderer, this, log);
            simThread = null;

            pauseButton.Enabled = false;
            stopButton.Enabled = false;
            nextStepButton.Enabled = false;
            nextStepBox.Enabled = false;

            renClickEvt = new MouseEventHandler(renderSurfaceClicked);

            simulator.stepsPerSec = float.Parse(speedBox.Text);
            updateSpeedLabel();
        }

        protected override void OnFormClosing (FormClosingEventArgs e)
        {
            if (simThread != null && simThread.IsAlive)
            {
                simThread.Abort();
            }

            base.OnFormClosing(e);
        }

        public void runButtonClicked (object sender, EventArgs args)
        {
            // this should never happen, but still...
            if (simThread != null && simThread.IsAlive)
            {
                return;
            }

            runButton.Enabled = false;

            string type = (string)testType.SelectedItem;
            simThread = new Thread(() => run(type));
            simThread.Name = "simThread";
            simThread.Start();
        }

        public void testTypeChanged (object sender, EventArgs args)
        {
            switch ((string)testType.SelectedItem)
            {
            case testTypeStrFile:
                testPath.Text = defTestPathMsg;
                testPath.Enabled = true;

                gridSizeLabel.Text = null;
                gridSizeBox.Enabled = false;

                nFlowsLabel.Text = null;
                nFlowsBox.Enabled = false;

                nBlockedLabel.Text = null;
                nBlockedBox.Enabled = false;
                break;
                
            case testTypeStrGen:
                testPath.Text = defTestSeedMsg;
                testPath.Enabled = true;

                gridSizeLabel.Text = "Grid size";
                gridSizeBox.Enabled = true;

                nFlowsLabel.Text = "Flows";
                nFlowsBox.Enabled = true;

                nBlockedLabel.Text = "Blocked";
                nBlockedBox.Enabled = true;
                break;
            };
        }

        public void speedCommitClicked (object sender, EventArgs args)
        {
            float newSpeed;
            if (!float.TryParse(speedBox.Text, out newSpeed))
            {
                log.print("Invalid format, number expected\n", Color.Red);
                return;
            }
            if (newSpeed < 0.1f || newSpeed > 1000.0f)
            {
                log.print("Invalid value, speed must be between 0.1 and 1000\n", Color.Red);
                return;
            }

            simulator.stepsPerSec = newSpeed;
            updateSpeedLabel();
        }

        public void pauseButtonClicked (object sender, EventArgs args)
        {
            setPaused(!simulator.paused);
        }

        public void stopButtonClicked (object sender, EventArgs args)
        {
            // the thread should always be running to be able to click the button
            if (simThread != null && simThread.IsAlive)
            {
                simThread.Abort();
            }

            renderSurface.Image = null;

            setPaused(true);
            setFormControlEnabled(runButton, true);
            setFormControlEnabled(pauseButton, false);
            setFormControlEnabled(stopButton, false);
            setFormControlEnabled(nextStepButton, false);
            setFormControlEnabled(nextStepBox, false);

            renderSurface.MouseClick -= renClickEvt;
            setFormControlText(simInfoSelPoint, null);
        }

        public void nextStepButtonClicked (object sender, EventArgs args)
        {
            simulator.addSteps(nextStepBox.Text);
            setFormControlText(simInfoSelPoint, null);
        }

        public void renderSurfaceClicked (object sender, MouseEventArgs args)
        {
            ClickedCell cell = simulator.getClickedCell(args.X, args.Y);

            if (cell.color == 0)
            {
                setFormControlText(simInfoSelPoint, string.Format("Cell: ({0}, {1})    Empty\n", cell.y, cell.x));
            }
            else if (cell.color == -1)
            {
                setFormControlText(simInfoSelPoint, string.Format("Cell: ({0}, {1})    Base\n", cell.y, cell.x));
            }
            else
            {
                setFormControlText(simInfoSelPoint, string.Format("Cell: ({0}, {1})    Flow: {2}\n", cell.y, cell.x, cell.color));
            }
        }

        public void updateInfoLabels (int stepsDone, int score)
        {
            setFormControlText(simInfoCurStep, string.Format("Current step: {0}", stepsDone));
            setFormControlText(simInfoScore, string.Format("Score: {0}", score));
        }

        public void simulationDone ()
        {
            setPaused(true);
            setFormControlEnabled(pauseButton, false);
            setFormControlEnabled(nextStepButton, false);
            setFormControlEnabled(nextStepBox, false);

            log.print("Simulation Complete\n\n");
        }

        private void run (string type)
        {
            if (type == null)
            {
                log.print("Error: No test type selected\n", Color.Red);
            }
            else if (type == testTypeStrFile)
            {
                bool ret = 
                    executor.copyAndCheckInput(testPath.Text) && 
                    executor.runSolution(solPathBox.Text, runCommandBox.Text) && 
                    executor.checkOutput();
                if (ret)
                {
                    runSimulation();
                }
            }
            else if (type == testTypeStrGen)
            {
                bool ret = 
                    generateTest() && 
                    executor.runSolution(solPathBox.Text, runCommandBox.Text) && 
                    executor.checkOutput();
                if (ret)
                {
                    runSimulation();
                }
            }

            setFormControlEnabled(runButton, true);
        }

        private bool generateTest ()
        {
            int seed;

            if (testPath.Text == null || testPath.Text == "" || testPath.Text == defTestSeedMsg)
            {
                seed = -1;
            }
            else
            {
                if (!int.TryParse(testPath.Text, out seed))
                {
                    log.print("Error: Invalid seed (expected an integer)\n", Color.Red);
                    return false;
                }
            }

            int gridSz, nFlows, nBlocked;

            if (gridSizeBox.Text == null || !int.TryParse(gridSizeBox.Text, out gridSz))
            {
                log.print("Error: Invalid argument \"Grid Size\"\n", Color.Red);
                return false;
            }
            if (gridSz < 6 || gridSz > 100)
            {
                log.print("Error: Invalid value for argument \"Grid Size\", must be in range [6, 100]\n", Color.Red);
                return false;
            }

            if (nFlowsBox.Text == null || !int.TryParse(nFlowsBox.Text, out nFlows))
            {
                log.print("Error: Invalid argument \"Flows\"\n", Color.Red);
                return false;
            }
            if (nFlows < 0 || nFlows > 250)
            {
                log.print("Error: Invalid value for argument \"Flows\", must be in range [0, 250]\n", Color.Red);
                return false;
            }

            if (nBlockedBox.Text == null || !int.TryParse(nBlockedBox.Text, out nBlocked))
            {
                log.print("Error: Invalid argument \"Blocked\"\n", Color.Red);
                return false;
            }
            if (nBlocked < 0 || nBlocked > 500)
            {
                log.print("Error: Invalid value for argument \"Blocked\", must be in range [0, 500]\n", Color.Red);
                return false;
            }

            if (!executor.generate(ref seed, gridSz, nFlows, nBlocked))
            {
                return false;
            }

            setFormControlText(testPath, seed.ToString());
            return true;
        }

        private bool runSimulation ()
        {
            simulator.init(executor.inputDest, executor.outputDest);

            setFormControlEnabled(pauseButton, true);
            setPaused(true);
            setFormControlEnabled(stopButton, true);

            simulator.start();
            return true;
        }

        private void setPaused (bool paused)
        {
            simulator.paused = paused;

            setFormControlEnabled(nextStepButton, paused);
            setFormControlEnabled(nextStepBox, paused);
            updateSpeedLabel();

            if (paused)
            {
                // prevents adding the event twice
                renderSurface.MouseClick -= renClickEvt;
                renderSurface.MouseClick += renClickEvt;
            }
            else
            {
                renderSurface.MouseClick -= renClickEvt;
                setFormControlText(simInfoSelPoint, null);
            }
        }

        private void updateSpeedLabel ()
        {
            if (simThread == null || !simThread.IsAlive || simulator.paused)
            {
                setFormControlText(simInfoSpeed, string.Format("Speed: {0} <paused>", simulator.stepsPerSec));
            }
            else
            {
                setFormControlText(simInfoSpeed, string.Format("Speed: {0}", simulator.stepsPerSec));
            }
        }

        private void setFormControlEnabled (Control control, bool enabled)
        {
            if (control.InvokeRequired)
            {
                control.Invoke(new SetFormControlEnabledDelegate(setFormControlEnabled), new object[]{control, enabled});
            }
            else
            {
                control.Enabled = enabled;
            }
        }

        private void setFormControlText (Control label, string text)
        {
            if (label.InvokeRequired)
            {
                label.Invoke(new SetFormControlTextDelegate(setFormControlText), new object[]{label, text});
            }
            else
            {
                label.Text = text;
            }
        }

        delegate void SetFormControlTextDelegate (Control label, string text);
        delegate void SetFormControlEnabledDelegate (Control control, bool enabled);

        private Log log;
        private Executor executor;
        private Simulator simulator;
        private Renderer renderer;
        private Thread simThread;

        private MouseEventHandler renClickEvt;
    }
}
