﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;

namespace s4_visualizer
{
    struct ClickedCell
    {
        public ClickedCell (int _x, int _y, int _color)
        {
            x = _x;
            y = _y;
            color = _color;
        }

        public int x;
        public int y;
        public int color;
    }

    class Simulator
    {
        public Simulator(Renderer _renderer, Form1 _mainForm , Log _log)
        {
            renderer = _renderer;
            mainForm = _mainForm;
            log = _log;
            
            paused = false;
            stepsPerSec = 1.0f;
            stepsQueued = 0;
            drawInterval = 0.25f;
        }

        public void init(string inputFile, string outputFile)
        {
            readInput(inputFile);

            readOutput(outputFile);
            
            // draw the initial state of the grid
            renderer.draw(grid, gridSz);
        }

        public void start()
        {
            paused = true;
            stepsQueued = 0;
            stepsDone = 0;
            curColor = 0;
            untilColor = curColor;
            score = 0;
            pointsFilled = 0;
            flowsConnected = 0;
            curConnected = false;
            mainForm.updateInfoLabels(stepsDone, score);

            DateTime prevStepTime = DateTime.Now;
            double timeSinceLastDraw = 0.0f;

            // find the first command
            curCommand = commands[curColor].First;
            while (curCommand == null)
            {
                curColor++;
                if (curColor == nPoints)
                    break;

                curCommand = commands[curColor].First;
            }

            while (curColor < nPoints)
            {
                if (!paused)
                {
                    nextStep();
                    Thread.Sleep((int)(1000.0f / stepsPerSec)); // in miliseconds
                }
                else if (stepsQueued > 0 || untilColor > curColor)
                {
                    nextStep();

                    if (stepsQueued > 0)
                        Interlocked.Decrement(ref stepsQueued);
                }
                else
                {
                    Thread.Sleep(100); // in miliseconds
                }

                DateTime curTime = DateTime.Now;
                timeSinceLastDraw += (curTime - prevStepTime).TotalSeconds;

                if (timeSinceLastDraw > drawInterval)
                {
                    renderer.draw(grid, gridSz);
                    mainForm.updateInfoLabels(stepsDone, score);

                    if (timeSinceLastDraw > 2 * drawInterval)
                    {
                        timeSinceLastDraw = 0.0;
                    }
                    else
                    {
                        timeSinceLastDraw -= drawInterval;
                    }
                }

                prevStepTime = DateTime.Now;
            }

            renderer.draw(grid, gridSz);
            mainForm.updateInfoLabels(stepsDone, score);
            mainForm.simulationDone();
        }

        public void addSteps(string steps)
        {
            if (steps == null || steps == "")
            {
                Interlocked.Increment(ref stepsQueued);
            }
            else if (steps == "f")
            {
                untilColor = curColor + 1;
            }
            else if (steps == "e")
            {
                untilColor = nPoints + 1;
            }
            else
            {
                int advance;

                if (!int.TryParse(steps, out advance) || advance < 0)
                {
                    log.print("Error: Invalid argument\n", Color.Red);
                }
                else
                {
                    Interlocked.Add(ref stepsQueued, advance);
                }
            }
        }

        public ClickedCell getClickedCell (int mouseX, int mouseY)
        {
            Point cell = renderer.getClickedCell(mouseX, mouseY, gridSz);
            int color = grid[cell.X, cell.Y];

            return new ClickedCell(cell.X, cell.Y, color);
        }

        private void readInput (string inputFile)
        {
            StreamReader sr = new StreamReader(inputFile);
            int indx = 0;
            string[] split = sr.ReadToEnd().Split(new char[] { ' ', '\t', '\n', '\r' });
            sr.Close();

            List<string> content = new List<string>();
            for (int i = 0; i < split.Length; i++)
            {
                if (split[i].Length >= 1)
                {
                    content.Add(split[i]);
                }
            }
            indx = 0;

            gridSz = int.Parse(content[indx++]);
            grid = new int[gridSz, gridSz];

            nPoints = int.Parse(content[indx++]);

            for (int i = 0; i < nPoints; i++)
            {
                int x, y;
                y = int.Parse(content[indx++]);
                x = int.Parse(content[indx++]);
                grid[x, y] = i + 1;

                y = int.Parse(content[indx++]);
                x = int.Parse(content[indx++]);
                grid[x, y] = i + 1;
            }

            int nBlocked = int.Parse(content[indx++]);
            for (int i = 0; i < nBlocked; i++)
            {
                int x, y;
                y = int.Parse(content[indx++]);
                x = int.Parse(content[indx++]);
                grid[x, y] = -1;
            }
        }

        private bool readOutput (string outputFile)
        {
            StreamReader sr = new StreamReader(outputFile);
            int indx = 0;
            string[] split = sr.ReadToEnd().Split(new char[] { ' ', '\t', '\n', '\r' });
            sr.Close();

            List<string> content = new List<string>();
            for (int i = 0; i < split.Length; i++)
            {
                if (split[i].Length >= 1)
                {
                    content.Add(split[i]);
                }
            }
            indx = 0;

            commands = new LinkedList<Point>[nPoints];

            for (int i = 0; i < nPoints; i++)
            {
                int nCom;
                commands[i] = new LinkedList<Point>();
                
                nCom = int.Parse(content[indx++]);
                
                for (int c = 0; c < nCom; c++)
                {
                    int x, y;
                    y = int.Parse(content[indx++]);
                    x = int.Parse(content[indx++]);;
                    commands[i].AddLast(new Point(x, y));
                }
            }

            return true;
        }

        private void nextStep()
        {
            Point com = curCommand.Value;
            curCommand = curCommand.Next;

            grid[com.X, com.Y] = curColor + 1;
            curConnected = true;

            while (curCommand == null)
            {
                if (curConnected)
                {
                    pointsFilled += commands[curColor].Count;
                    flowsConnected++;
                    calcScore();
                    curConnected = false;
                }

                curColor++;
                if (curColor == nPoints)
                    break;

                curCommand = commands[curColor].First;
            }

            stepsDone++;
        }

        private void calcScore ()
        {
            score = flowsConnected * pointsFilled;
        }

        delegate void UpdateLabelDelegate ();

        private int[,] grid;
        private int gridSz;
        private int nPoints;
        private LinkedList<Point>[] commands;
        private int curColor;
        private LinkedListNode<Point> curCommand;

        public int score;
        private int pointsFilled;
        private int flowsConnected;
        private bool curConnected;

        public bool paused;
        public float stepsPerSec;
        private int stepsQueued;
        private int untilColor;
        private int stepsDone;
        private double drawInterval;

        private Renderer renderer;
        private Form1 mainForm;
        private Log log;
    };
}